struct Page_Ppm
{
  GtkWidget *frame_ppm;
  GtkWidget *label_ppm;
  GtkWidget *radio_raw_ppm;
  GtkWidget *radio_ascii_ppm;
  GtkWidget *vbox_ppm;
  GSList *group_ppm;
};

struct Page_Jpeg 
{
  GtkWidget *frame_jpeg;
  GtkWidget *vbox_jpeg;
  GtkWidget *table_jpeg;
  GtkWidget *quality_jpeg;
  GtkWidget *smooth_jpeg;
  GtkWidget *label_jpeg;
  GtkWidget *optimize_jpeg;
  GtkObject *quality_adj_jpeg;
  GtkObject *smooth_adj_jpeg;
};

enum Save_Types {
  ppm = 0,
  jpeg = 1
};

void save_dialog();
void file_select(GtkWidget *widget, GdkEvent *event, gpointer data);
