#include <linux/videodev.h>

#define TIMEOUT_LEN 1000
#define MAX_RGBA_IMAGE_SIZE ((704*576*4)+256)

void delete_event( GtkWidget *widow, GdkEvent  *event, gpointer   data );
static gint configure_event (GtkWidget *widget, GdkEventConfigure *event);
static gint expose_event (GtkWidget *widget, GdkEventExpose *event);
gint grab (gpointer data);
void clearimg();
void freeze(GtkWidget *widget, GdkEvent  *event, gpointer   data);
void setbrightness(GtkAdjustment *adj, GtkWidget *none);
void setwhite(GtkAdjustment *adj, GtkWidget *none);
void setcontrast(GtkAdjustment *adj, GtkWidget *none);
void setsizefull( GtkWidget *widget, GdkEvent  *event, gpointer   data );
void setsizehalf( GtkWidget *widget, GdkEvent  *event, gpointer   data );
void setsizequarter( GtkWidget *widget, GdkEvent  *event, gpointer data);
void setdepth4(GtkWidget *widget, GdkEvent *event, gpointer data);
void setdepth6(GtkWidget *widget, GdkEvent *event, gpointer data);
void snap(GtkWidget *widget, GdkEvent *event, gpointer data);

extern struct video_window vid_win;
extern unsigned char pic[MAX_RGBA_IMAGE_SIZE];
extern int frozen;
extern guint timeoutid;
